/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <amqpcpp.h>
#include <amqpcpp/libevent.h>

#include <event2/event.h>

#include <algorithm>
#include <chrono>
#include <cstdlib>
#include <exception>
#include <functional>
#include <iomanip>
#include <iostream>
#include <list>
#include <map>
#include <memory>
#include <sstream>
#include <thread>
#include <utility>

const auto t0 = std::chrono::steady_clock::now();

auto current_time()
{
    const auto us = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::steady_clock::now() - t0);
    std::ostringstream out;
    out << (us.count() / 1'000'000) << "." << std::setw(6) << std::setfill('0') << (us.count() % 1'000'000);
    return out.str();
}

#define LOG(MSG) do { \
    std::cerr << ::current_time() << " - " << MSG << " at: " << __LINE__ << std::endl; \
} while(false)

class Timer
{
public:
    explicit Timer();
    ~Timer();
    using TimePoint = std::chrono::steady_clock::time_point;
    using Callback = std::function<void(TimePoint)>;
    using CallbackList = std::list<Callback>;
    struct CallbackHandle
    {
        TimePoint scheduled;
        CallbackList::const_iterator callback_list_iter;
    };
    CallbackHandle register_callback(TimePoint scheduled, Callback callback);
    void erase_callback(const CallbackHandle& callback_handle);
    void loop();
    struct event_base* service() const noexcept;
private:
    bool is_stopped() const noexcept;
    using CallbacksByTime = std::map<TimePoint, CallbackList>;
    std::unique_ptr<struct event_base, void(*)(struct event_base*)> service_;
    CallbacksByTime scheduler_;
};

class Client
{
public:
    explicit Client(Timer& timer, const AMQP::Address& address);
    void onAttached(AMQP::TcpConnection*);
    void onConnected(AMQP::TcpConnection*);
    bool onSecured(AMQP::TcpConnection*, const SSL*);
    void onProperties(AMQP::TcpConnection*, const AMQP::Table&, AMQP::Table&);
    void onReady(AMQP::TcpConnection*);
    void onError(AMQP::TcpConnection*, const char *message);
    void onLost(AMQP::TcpConnection*);
    uint16_t onNegotiate(AMQP::TcpConnection*, uint16_t);
    void onHeartbeat(AMQP::TcpConnection*);
    void onClosed(AMQP::TcpConnection*);
    void onDetached(AMQP::TcpConnection*);
    struct event_base* service();
private:
    void on_tick_7(const Timer::TimePoint& now);
    void on_tick_9(const Timer::TimePoint& now);
    Timer& timer_;
    AMQP::Address address_;
    std::unique_ptr<AMQP::TcpConnection> conn_;
    AMQP::TcpConnection* broken_conn_;
};

int main()
{
    Timer timer;
    const Client client{timer, AMQP::Address{"localhost", 6672, {"guest", "guest"}, "/"}};
    timer.loop();
    return EXIT_SUCCESS;
}

std::unique_ptr<AMQP::TcpConnection> make_tcp_connection(
        std::unique_ptr<AMQP::TcpHandler> io_handler,
        const AMQP::Address& address);

std::unique_ptr<AMQP::TcpHandler> make_boost_asio_handler(Client& client);

Client::Client(Timer& timer, const AMQP::Address& address)
    : timer_{timer},
      address_{address},
      conn_{make_tcp_connection(make_boost_asio_handler(*this), address)},
      broken_conn_{nullptr}
{
    on_tick_7(Timer::TimePoint::clock::now());
    on_tick_9(Timer::TimePoint::clock::now());
}

void Client::onAttached(AMQP::TcpConnection*)
{
    LOG("onAttached");
}

void Client::onConnected(AMQP::TcpConnection*)
{
    LOG("onConnected");
}

bool Client::onSecured(AMQP::TcpConnection*, const SSL*)
{
    LOG("onSecured");
    // default implementation: do not inspect anything, just allow the connection
    return true;
}

void Client::onProperties(AMQP::TcpConnection*, const AMQP::Table&, AMQP::Table&)
{
    LOG("onProperties");
}

void Client::onReady(AMQP::TcpConnection*)
{
    LOG("onReady");
}

void Client::onError(AMQP::TcpConnection* connection, const char *message)
{
    LOG("onError(\"" << message << "\") " << (connection == conn_.get() ? "current" : "old"));
    if (connection == conn_.get())
    {
        broken_conn_ = connection;
    }
}

void Client::onLost(AMQP::TcpConnection*)
{
    LOG("onLost");
}

uint16_t Client::onNegotiate(AMQP::TcpConnection*, uint16_t)
{
    LOG("onNegotiate");
    return 0;//heartbeat turned off
}

void Client::onHeartbeat(AMQP::TcpConnection*)
{
    LOG("onHeartbeat");
}

void Client::onClosed(AMQP::TcpConnection*)
{
    LOG("onClosed");
}

void Client::onDetached(AMQP::TcpConnection*)
{
    LOG("onDetached");
    if (broken_conn_ != nullptr)
    {
        LOG("try_reconnect");
        conn_.reset();
        std::this_thread::sleep_for(std::chrono::seconds{5});
        conn_ = make_tcp_connection(make_boost_asio_handler(*this), address_);
        broken_conn_ = nullptr;
    }
}

void Client::on_tick_7(const Timer::TimePoint& now)
{
    LOG("tick with period 7");
    timer_.register_callback(now + std::chrono::seconds{7}, [this](auto&& now) { on_tick_7(now); });
}

void Client::on_tick_9(const Timer::TimePoint& now)
{
    LOG("tick with period 9");
    timer_.register_callback(now + std::chrono::seconds{9}, [this](auto&& now) { on_tick_9(now); });
}

struct event_base* Client::service()
{
    return timer_.service();
}

std::unique_ptr<AMQP::TcpConnection> make_tcp_connection(
        std::unique_ptr<AMQP::TcpHandler> io_handler,
        const AMQP::Address& address)
{
    using TcpHandlerStorage = std::unique_ptr<AMQP::TcpHandler>;
    class Connection : private TcpHandlerStorage, public AMQP::TcpConnection
    {
    public:
        explicit Connection(
                std::unique_ptr<AMQP::TcpHandler> io_handler,
                const AMQP::Address& address)
            : TcpHandlerStorage{std::move(io_handler)},
              AMQP::TcpConnection{this->TcpHandlerStorage::get(), address}
        { }
        virtual ~Connection() override
        { }
    };
    LOG("make_tcp_connection");
    return std::make_unique<Connection>(std::move(io_handler), address);
}

std::unique_ptr<AMQP::TcpHandler> make_boost_asio_handler(Client& client)
{
    class IoHandler : public AMQP::LibEventHandler
    {
    public:
        IoHandler(Client& client)
            : AMQP::LibEventHandler{client.service()},
              client_{client}
        { }
    private:
        /**
         *  Method that is called immediately after a connection has been constructed. 
         *  @param  connection      The connection object that was just constructed
         */
        void onAttached(AMQP::TcpConnection* connection) override
        {
            client_.onAttached(connection);
        }
        /**
         *  Method that is called when the TCP connection ends up in a connected state
         *  This method is called after the TCP connection has been set up, but before
         *  the (optional) secure TLS connection is ready, and before the AMQP login 
         *  handshake has been completed. If this step has been set, the onLost()
         *  method will also always be called when the connection is closed.
         *  @param  connection  The TCP connection
         */
        void onConnected(AMQP::TcpConnection* connection) override
        {
            client_.onConnected(connection);
        }
        /**
         *  Method that is called after a TCP connection has been set up and the initial 
         *  TLS handshake is finished too, but right before the AMQP login handshake is
         *  going to take place and the first data is going to be sent over the connection. 
         *  This method allows you to inspect the TLS certificate and other connection 
         *  properties, and to break up the connection if you find it not secure enough. 
         *  The default implementation considers all connections to be secure, even if the 
         *  connection has a self-signed or even invalid certificate. To be more strict, 
         *  override this method, inspect the certificate and return false if you do not 
         *  want to use the connection. The passed in SSL pointer is a pointer to a SSL 
         *  structure from the openssl library. This method is only called for secure 
         *  connections (connection with an amqps:// address).
         *  @param  connection      The connection for which TLS was just started
         *  @param  ssl             Pointer to the SSL structure that can be inspected
         *  @return bool            True to proceed / accept the connection, false to break up
         */
        bool onSecured(AMQP::TcpConnection* connection, const SSL* ssl) override
        {
            return client_.onSecured(connection, ssl);
        }
        /**
         *  Method that is called when the RabbitMQ server and your client application  
         *  exchange some properties that describe their identity.
         *  @param  connection      The connection about which information is exchanged
         *  @param  server          Properties sent by the server
         *  @param  client          Properties that are to be sent back
         */
        void onProperties(AMQP::TcpConnection* connection, const AMQP::Table& server, AMQP::Table& client) override
        {
            client_.onProperties(connection, server, client);
        }
        /**
         *  Method that is called after the AMQP login handshake has been completed
         *  and the connection object is ready for sending out actual AMQP instructions
         *  @param  connection  The TCP connection
         */
        void onReady(AMQP::TcpConnection* connection) override
        {
            client_.onReady(connection);
        }
        /**
         *  Method that is called when the connection ends up in an error state
         *  This could either be an error at the AMQP level, but could also
         *  be an error at the TCP of SSL level (like a broken connection).
         *  If the connection is connected (the onConnected() method was called
         *  before), the onLost() method is going to be called too.
         *  @param  connection  The TCP connection
         *  @param  message     Error message
         */
        void onError(AMQP::TcpConnection* connection, const char *message) override
        {
            client_.onError(connection, message);
        }
        /**
         *  Method that is called when the TCP connection is lost or closed. This
         *  is always called if you have also received a call to onConnected().
         *  @param  connection  The TCP connection
         */
        void onLost(AMQP::TcpConnection* connection) override
        {
            client_.onLost(connection);
        }
        /**
         *  Method that is called when the heartbeat frequency is negotiated
         *  between the server and the client. Applications can override this method
         *  if they want to use a different heartbeat interval (for example: return 0
         *  to disable heartbeats)
         *  @param  connection      The connection that suggested a heartbeat interval
         *  @param  interval        The suggested interval from the server
         *  @return uint16_t        The interval to use
         *
         *  @see ConnectionHandler::onNegotiate
         */
        uint16_t onNegotiate(AMQP::TcpConnection* connection, uint16_t interval) override
        {
            return client_.onNegotiate(connection, interval);
        }
        /**
         *  Method that is called when the server sends a heartbeat to the client
         *  @param  connection      The connection over which the heartbeat was received
         *  @see    ConnectionHandler::onHeartbeat
         */
        void onHeartbeat(AMQP::TcpConnection* connection) override
        {
            client_.onHeartbeat(connection);
        }
        /**
         *  Method that is called when the AMQP protocol was gracefully ended. 
         *  This is the counter-part of a call to connection.close(). Note that
         *  the underlying TCP connection is still alive, and onLost() and
         *  onDetached() (see below) are going to be called too.
         *  @param  connection  The TCP connection
         */
        void onClosed(AMQP::TcpConnection* connection) override
        {
            client_.onClosed(connection);
        }
        /**
         *  Method that is called when the handler will no longer be notified.
         *  This is the last call to your handler, and it is typically used
         *  to clean up stuff.
         *  @param  connection      The connection that is being destructed
         */
        void onDetached(AMQP::TcpConnection* connection) override
        {
            client_.onDetached(connection);
        }
        Client& client_;
    };
    return std::make_unique<IoHandler>(client);
}

Timer::Timer()
    : service_{event_base_new(), event_base_free}
{ }

Timer::~Timer()
{
    try
    {
        LOG(scheduler_.size() << " timeouts rest");
    }
    catch (...) { }
}

Timer::CallbackHandle Timer::register_callback(TimePoint scheduled, Callback callback)
{
//    LOG("register callback function triggered by time limited polling loop");
    auto& callback_list = scheduler_[scheduled];
    return CallbackHandle{scheduled, callback_list.insert(end(callback_list), std::move(callback))};
}

void Timer::erase_callback(const CallbackHandle& callback_handle)
{
    auto scheduler_iter = scheduler_.find(callback_handle.scheduled);
    if (scheduler_iter != end(scheduler_))
    {
        scheduler_iter->second.erase(callback_handle.callback_list_iter);
        if (scheduler_iter->second.empty())
        {
            scheduler_.erase(scheduler_iter);
        }
    }
}

template <typename Rep, typename Period>
struct ::timeval to_timeval(const std::chrono::duration<Rep, Period>& d)
{
    const auto us = std::chrono::duration_cast<std::chrono::microseconds>(d).count();
    struct ::timeval timeval = {0, 0};
    if (0 < us)
    {
        timeval.tv_sec = us / 1'000'000;
        timeval.tv_usec = us % 1'000'000;
    }
    return timeval;
}

void Timer::loop()
{
    while (!is_stopped())
    {
        if (scheduler_.empty())
        {
            LOG("run infinity polling loop");
            event_base_loop(service(), 0);
            LOG("infinity polling loop done");
            continue;
        }
        const auto dt = to_timeval(begin(scheduler_)->first - TimePoint::clock::now());
        const auto loopexit_code = event_base_loopexit(service(), &dt);
        LOG("loop exit scheduled in " << dt.tv_sec << "." << std::setw(6) << std::setfill('0') << dt.tv_usec << (loopexit_code == 0 ? "" : ", error"));
        const auto loop_code = event_base_loop(service(), 0);
        LOG("loop exited with code " << loop_code);
        const auto now = TimePoint::clock::now();
        const auto first_in_the_future = std::find_if(begin(scheduler_), end(scheduler_), [&](auto&& scheduled)
        {
            if (now < scheduled.first)
            {
                return true;
            }
            std::for_each(begin(scheduled.second), end(scheduled.second), [&scheduled](auto&& callback)
            {
                try
                {
                    callback(scheduled.first);
                }
                catch (const std::exception& e)
                {
                    LOG("scheduled callback failure: " << e.what());
                }
                catch (...)
                {
                    LOG("scheduled callback failure");
                }
            });
            return false;
        });
        scheduler_.erase(begin(scheduler_), first_in_the_future);
    }
}

struct event_base* Timer::service() const noexcept
{
    return service_.get();
}

bool Timer::is_stopped() const noexcept
{
    return (service_ == nullptr) ||
           event_base_got_break(service());
}
